package org.springblade.useranalysis.service.normal;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import org.apache.commons.lang3.StringUtils;
import org.springblade.useranalysis.common.utils.HashKit;
import org.springblade.useranalysis.common.utils.JsonMap;
import org.springblade.useranalysis.dto.param.ExcelExportParam;
import org.springblade.useranalysis.entity.TagColumn;
import org.springblade.useranalysis.entity.UserGroupEntity;
import org.springblade.useranalysis.entity.UserProfileExcelLog;
import org.springblade.useranalysis.entity.clickhouse.ZUserTagEntity;
import org.springblade.useranalysis.service.UserProfileService;
import org.springblade.useranalysis.service.UserTagsService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import com.alibaba.excel.EasyExcel;
import com.alibaba.excel.ExcelWriter;
import com.alibaba.excel.converters.longconverter.LongStringConverter;
import com.alibaba.excel.util.FileUtils;
import com.alibaba.fastjson2.JSON;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.github.xiaoymin.knife4j.core.util.StrUtil;

import lombok.extern.slf4j.Slf4j;

/**
 * @author william
 *
 */
@Slf4j
@Service
public class ExcelService {

	private final static int PAGE_SIZE = 50000;

	@Autowired
	private UserProfileExcelLogService excelLogService;

	@Autowired
	private UserTagsService userTagsService;

	@Autowired
	private UserProfileService userProfileService;
	
	@Autowired
	private UserGroupRuleService userGroupRuleService;

	final ExecutorService threadPool = new ThreadPoolExecutor(1, 2, 5, TimeUnit.MINUTES,
			// 有界队列5000，超过队列长度，无异常抛弃
			new ArrayBlockingQueue<>(5000), new ThreadFactory() {
				@Override
				public Thread newThread(Runnable r) {
					Thread thread = new Thread(r);
					thread.setDaemon(false);
					thread.setName("Export-User-Group-Pool-");
					return thread;
				}
			}, new ThreadPoolExecutor.DiscardPolicy());

	public void excute(ExcelExportParam params, UserProfileExcelLog excelLog, UserGroupEntity userGroup) {
		threadPool.execute(() -> {
			log.info(excelLog.getId() + " - exporting......");
			long start = System.currentTimeMillis();
			try {
				String groupId = params.getGroupId();
				// 使用前端传递的日期范围
//				String monthStart = params.getMonthStart();
//				String monthEnd = params.getMonthEnd();

				// 使用圈定人群的日期范围
//				String monthStart = "";
//				String monthEnd = "";
//				JSONObject rule = JSON.parseObject(userGroup.getRules());
//				if (rule != null) {
//					if (rule.containsKey("daterange")) {
//						JSONArray daterange = rule.getJSONArray("daterange");
//						if (daterange != null && daterange.size() == 2) {
//							monthStart = daterange.getString(0);
//							monthEnd = daterange.getString(1);
//						}
//					}
//				}
				QueryWrapper<ZUserTagEntity> whereWrapper = userGroupRuleService.parse(userGroup);
				whereWrapper.eq("ur.user_group_id ", groupId);
				//whereWrapper解析后已经包含日期范围，所以不需要再添加时间，只需要追加人群id即可
//				whereWrapper.ge("tag.THEMONTH_V", monthStart);
//				whereWrapper.le("tag.THEMONTH_V", monthEnd);
				ConcurrentHashMap<String, String> mobileMap = new ConcurrentHashMap<>();
				boolean isSuccess = handle(mobileMap, params, excelLog, userGroup, whereWrapper);
				if (isSuccess) {
					excelLog.setCount(params.getCount());
					excelLog.setStatus(UserProfileExcelLogService.STATUS_SUCCESS);
					excelLogService.saveOrUpdate(excelLog);
					log.info(excelLog.getId() + " - export success");
				} else {
					log.info(excelLog.getId() + " - export failure");
					deleteCache(excelLog);
				}
			} catch (Exception e) {
				deleteCache(excelLog);
				log.info(excelLog.getId() + " - export failure,the error is ：" + e.getMessage());
			}
			long end = System.currentTimeMillis();
			log.info(excelLog.getId() + " - take up time is " + (end - start) / 1000.f + "second");
		});
	}

	private void deleteCache(UserProfileExcelLog excelLog) {
		try {
			// 导出失败
			excelLog.setStatus(UserProfileExcelLogService.STATUS_FAILURE);
			excelLogService.saveOrUpdate(excelLog);
			// 删除导出失败的垃圾文件
			FileUtils.delete(new File(excelLog.getFilePath()));
		} catch (Exception e2) {
			log.info(excelLog.getId() + " - delete the file that export failure,the error is " + e2.getMessage());
		}
	}

	private boolean handle(ConcurrentHashMap<String, String> mobileMap, ExcelExportParam params,
			UserProfileExcelLog excelLog, UserGroupEntity userGroup, QueryWrapper<ZUserTagEntity> whereWrapper) {
		boolean isSuccess = true;
		List<String> codeList = params.getTagList();
		boolean desens = params.isDesens();
		StringBuffer fields = new StringBuffer();
		List<TagColumn> columnList = getFieldList(codeList);
		columnList.forEach(col -> {
			if (StrUtil.isNotBlank(col.columnV) && !col.columnV.contains("USERID")) {
				if (StrUtil.isNotBlank(col.showHistory) && col.showHistory.equalsIgnoreCase("1")) {
					fields.append(col.columnV);
				} else {
					fields.append(col.columnC);
				}
				fields.append(",");
			}
		});
		fields.append("USERID_V,THEMONTH_V");
		log.info("export fields is :" + fields.toString());
		File excelFile = new File(excelLog.getFilePath());
		ExcelWriter excelWriter = EasyExcel.write(excelFile)
				.registerConverter(new LongStringConverter())
				.head(buidlHeader(columnList))
				.build();
		// 需要导出的数量
		int count = params.getCount();
		log.info("export count need is :" + count);

		// 已经导出的数量
		int exportedCountTotal = 0;

		// 下次查询偏移量
		int offset = 0;

		while (true) {
			int pageSize = PAGE_SIZE;
			String sqlLimit = "limit " + offset + "," + pageSize;
			List<Map<String, Object>> tagList = null;
			try {
				log.info("export limit sql is :" + sqlLimit);
				tagList = userTagsService.exportUserGroup(fields.toString(), whereWrapper, sqlLimit);
			} catch (Exception e) {
				log.error("ExcelService handle fail, the error is " + e.getMessage());
				isSuccess = false;
				break;
			}

			if (tagList == null || tagList.isEmpty()) {
				// 没有数据
				break;
			}

			// 本次去重后导出的数量
			int exportedCountCur = excel(mobileMap, desens, excelWriter, columnList, tagList, excelLog.getGroupName());
			exportedCountTotal += exportedCountCur;
			if (exportedCountTotal == count || tagList.size() < pageSize) {
				// 导出数量完成 或者 已经查询完成
				log.info("export exportedCountTotal is :" + exportedCountTotal);
				log.info("export tagList.size() is :" + tagList.size());
				break;
			}

			offset += pageSize;
		}

		if (excelWriter != null) {
			excelWriter.finish();
		}

		log.info("export count total is :" + exportedCountTotal);
		params.setCount(exportedCountTotal);
		return isSuccess;
	}

	/**
	 * 结果导出到excel文件中
	 * 
	 * @param desens      手机号是否脱敏
	 * @param excelWriter excel写入器
	 * @param columnList  字段
	 * @param tagList     结果
	 * @param groupName   人群名称
	 * 
	 * @return 本次写入的去重后的手机号数量
	 */
	private int excel(ConcurrentHashMap<String, String> mobileMap, boolean desens, ExcelWriter excelWriter,
			List<TagColumn> columnList,
			List<Map<String, Object>> tagList, String groupName) {
		int count = 0;
		groupName = StringUtils.isNotBlank(groupName) ? groupName : "1";
		List<List<Object>> list = new ArrayList<List<Object>>();
		for (Map<String, Object> tag : tagList) {
			List<Object> data = new ArrayList<Object>();
			/** ---------去重手机号------ **/
			String mobile = tag.get("USERID_V").toString();
			if (mobileMap.containsKey(mobile)) {
				continue;
			}
			mobileMap.put(mobile, mobile);
			count += 1;
			/** ---------去重手机号------ **/
			if (desens) {
				// 手机号脱敏
				data.add(HashKit.sha1(tag.get("USERID_V").toString()));
			} else {
				data.add(tag.get("USERID_V"));
			}
			Object month = tag.get("THEMONTH_V");
			data.add(month != null ? month.toString() : "");

			columnList.forEach(col -> {
				Object obj = tag.get(col.columnC);
				if (StrUtil.isNotBlank(col.showHistory) && col.showHistory.equalsIgnoreCase("1")) {
					obj = tag.get(col.columnV);
				}
				if (StrUtil.isNotBlank(col.columnType) && col.columnType.equalsIgnoreCase("array")) {
					String value = "NaN";
					JsonMap jsonMap = JsonMap.parseJson(JSON.toJSONString(tag));
					String[] array = col.columnC.split("\\.");
					if (array != null && array.length == 2 && jsonMap != null) {
						String arrayParent = array[0];
						String arrayChild = array[1];
						JsonMap parent = jsonMap.getMap(arrayParent);
						if (parent != null) {
							List<String> childList = parent.getStringList(arrayChild);
							if (childList != null && !childList.isEmpty()) {
								value = JsonMap.toJson(childList);
							}
						}
					} else {
						value = obj != null ? obj.toString() : "NaN";
					}
					data.add(value);
				} else {
					data.add(obj != null ? obj.toString() : "NaN");
				}
			});

			list.add(data);
		}

		excelWriter.write(list, EasyExcel.writerSheet(groupName).build());
		return count;
	}

	/**
	 * 动态生成表头
	 */
	private List<List<String>> buidlHeader(List<TagColumn> columnList) {
		List<List<String>> list = new ArrayList<List<String>>();

		List<String> headPhone = new ArrayList<String>();
		headPhone.add("手机号");
		list.add(headPhone);

		List<String> headMonth = new ArrayList<String>();
		headMonth.add("月份");
		list.add(headMonth);

		columnList.forEach(col -> {
			List<String> head = new ArrayList<String>();
			head.add(col.cnName);
			list.add(head);
		});
		return list;
	}

	private List<TagColumn> getFieldList(List<String> codeList) {
		List<TagColumn> list = new ArrayList<>();
		if (codeList != null && !codeList.isEmpty()) {
			codeList.forEach(code -> {
				TagColumn weiduColumn = userProfileService.getColumNameExport(code);
				if (weiduColumn != null && weiduColumn.columnC != null
						&& !weiduColumn.columnC.contains("EVENT_UNSPNAME")) {
					// 该字段会严重影响查询速度，暂时屏蔽掉
					list.add(weiduColumn);
				}
			});
		}
		return list;
	}
}
